#!/usr/bin/env python
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect, logging, os, re, sys
from plumbum import FG, TF, local
from plumbum.machines.paramiko_machine import ParamikoMachine

__all__ = ["ExpRun", "FreeBSD11", "FreeBSD12", "FreeBSD13"]

tbody_template = """
<tr id="{instance}_stats">
    <td><a href="{instance}/latest/">{jail_version_arch}</a></td>
    <td id="{instance}_status"></td>
    <td id="{instance}_stats_queued" class="queued"></td>
    <td id="{instance}_stats_built" class="built"></td>
    <td id="{instance}_stats_failed" class="failed"></td>
    <td id="{instance}_stats_skipped" class="skipped"></td>
    <td id="{instance}_stats_ignored" class="ignored"></td>
    <td id="{instance}_stats_remaining" class="remaining"></td>
    <td id="{instance}_elapsed"></td>
</tr>
"""


class Jail(object):
    def __init__(self, version, type, platform):
        self.version = version
        self.type = type
        self.platform = platform
    
    def __str__(self):
        return "%s-%s" % (self.version.replace(".", "_"), self.platform)
    
    @property
    def release(self):
        return "%s-%s" % (self.version, self.type)


class FreeBSD(object):
    def __init__(self, version, type):
        self.i386 = Jail(version, type, "i386")
        self.amd64 = Jail(version, type, "amd64")


FreeBSD11 = FreeBSD("11.2", "RELEASE")
FreeBSD12 = FreeBSD("12.0", "RELEASE")
FreeBSD13 = FreeBSD("13", "CURRENT")


class ExpRun(object):
    def __init__(self, jails):
        self._remote = ParamikoMachine("dbn.westeurope.cloudapp.azure.com", "root", connect_timeout=None)
        self.jails = jails
        self.env = {}
        
        try:
            self._kldload = self._remote["kldload"]["-n"]
            self._patch = self._remote["patch"]
            self._poudriere = self._remote["poudriere"]
            self._tmux = self._remote["tmux"]
            self._zfs = self._remote["zfs"]

            if not self._zfs["rollback", "-r", "poudriere/ports/default@clean"] & TF:
                self._poudriere("ports", "-c")
                self._zfs("snapshot", "poudriere/ports/default@clean")
            
            jails = [jail[0] for jail in (jail.split() for jail in self._poudriere("jail", "-l").split("\n")) if jail]
            for jail in self.jails:
                if str(jail) not in jails:
                    create_jail = self._poudriere["jail", "-c", "-j", jail, "-a", jail.platform, "-v", jail.release]
                    if jail.type == "CURRENT":
                        create_jail = create_jail["-m", "url=http://ftp.freebsd.org/pub/FreeBSD/snapshots/%s/%s/%s.0-%s" % (jail.platform, jail.platform, jail.version, jail.type)]
                    create_jail & FG
        except:
            self._remote.close()
            raise
   
    def apply_patch(self, path, patch_level = 1):
        with self._remote.tempdir() as r_tempdir:
            r_patchfile = r_tempdir / "%s.diff"
            r_patchfile.write(local.path(path).read())
            self._patch["-d", "/usr/local/poudriere/ports/default", "-Efsp%i" % patch_level, "-i", r_patchfile] & FG
        
    def close(self):
        self._remote.close()
    
    def exprun(self, name, ports, increment = False, allow_distcache = False):
        instances = []
        tbody = ""
        r_make_conf = self._remote.path("/usr/local/etc/poudriere.d/%s-make.conf" % name)
        if self.env:
            r_make_conf.write("\n".join("%s=%s" % (k, v) for k, v in self.env.items()))
        elif r_make_conf.exists():
            r_make_conf.delete()
        self._kldload("linux", "linprocfs", "nullfs", "tmpfs")
        for jail in self.jails:
            poudriere_cmd = self._poudriere["bulk", "-%s%st" % ("" if increment else "C", "" if allow_distcache else "F"), "-j", jail, '-z', name].__getitem__(ports)
            self._tmux["new", "-d", "%s || sh -i" % poudriere_cmd] & FG
            instances.append("%s-default-%s" % (jail, name))
            tbody += tbody_template.format(
                instance = instances[-1],
                jail_version_arch = "%s/%s" % (jail.version, jail.platform))
        data_dir = local.path(inspect.getfile(inspect.currentframe())).dirname
        html = (data_dir / "exp-run.html").read(encoding="utf8").format(
            asset_dir = "%s/latest" % instances[0],
            instances = "[%s]" % ', '.join('"%s"' % i for i in instances),
            name = name,
            tbody = tbody)
        r_data_dir = self._remote.path("/usr/local/poudriere/data/logs/bulk")
        (r_data_dir / ("%s.html" % name)).write(html, encoding="utf8")
        (r_data_dir / ("%s.js" % name)).write((data_dir / "exp-run.js").read())
        return "http://dbn.westeurope.cloudapp.azure.com/%s.html" % name
    
    def update_jails(self):
        for jail in self.jails:
            self._poudriere["jail", "-u", "-j", jail] & FG

    def update_ports(self):
        self._zfs("destroy", "poudriere/ports/default@clean")
        self._poudriere["ports", "-u"] & FG
        self._zfs("snapshot", "poudriere/ports/default@clean")


def patch_plumbum():
    from plumbum.machines import SshMachine
    from plumbum.machines.paramiko_machine import ParamikoPopen
    from plumbum.machines.session import ShellSession

    shell_cmd = "/bin/sh"

    def patch_popen(self, args, ssh_opts = (), **kwargs):
        cmdline = []
        cmdline.extend(ssh_opts)
        cmdline.append(self._fqhost)
        if args:
            if hasattr(self, "env"):
                cmdline.extend(["cd", str(self.cwd), "&&"])
                envdelta = self.env.getdelta()
                if envdelta:
                    cmdline.append("env")
                    cmdline.extend("%s=%s" % (k, v) for k, v in envdelta.items())
            if isinstance(args, (tuple, list)):
                cmdline.extend(args)
            else:
                cmdline.append(args)
        return self._ssh_command[tuple(cmdline)].popen(**kwargs)

    def patch_session(self, isatty = False, term = "vt100", width = 80, height = 24, new_session = False):
        # new_session is ignored for ParamikoMachine
        trans = self._client.get_transport()
        trans.set_keepalive(self._keep_alive)
        chan = trans.open_session()
        if isatty:
            chan.get_pty(term, width, height)
            chan.set_combine_stderr(True)
        chan.exec_command(shell_cmd)
        stdin = chan.makefile('wb', -1)
        stdout = chan.makefile('rb', -1)
        stderr = chan.makefile_stderr('rb', -1)
        proc = ParamikoPopen(["<shell>"], stdin, stdout, stderr, self.custom_encoding)
        return ShellSession(proc, self.custom_encoding, isatty)

    SshMachine.popen = patch_popen
    ParamikoMachine.session = patch_session
patch_plumbum()

def main(args):
    logging.root.setLevel(logging.INFO)
    logging.root.addHandler(logging.StreamHandler(sys.stdout))

    if len(args) < 4:
        print("usage: (--increment) (--update-ports) (--allow-distcache) (-p #) name patch ports...")
        sys.exit(1)

    increment = False
    allow_distcache = False
    patch_level = 1
    exprun = ExpRun([FreeBSD11.i386, FreeBSD11.amd64, FreeBSD12.i386, FreeBSD12.amd64, FreeBSD13.i386, FreeBSD13.amd64])
    try:
        while True:
            if args[1] == "--increment":
                increment = True
                args.pop(1)
            elif args[1] == "--update-ports":
                exprun.update_ports()
                args.pop(1)
            elif args[1] == "--update-jails":
                exprun.update_jails()
                args.pop(1)
            elif args[1] == "--allow-distcache":
                allow_distcache = True
                args.pop(1)
            elif args[1] == "-p":
                args.pop(1)
                patch_level = int(args.pop(1))
            else:
                break

        name = args[1]
        patch = args[2]
        ports = args[3:]
        
        exprun.apply_patch(patch, patch_level)
        print(exprun.exprun(name, ports, increment, allow_distcache))
    finally:
        exprun.close()


if __name__ == "__main__":
    main(sys.argv)
